﻿using System;
using System.Collections;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using Core;

namespace EmbeddedWebserver.Core.Internal.Abstract
{
    public abstract class SocketServerBase : IDisposable
    {
        #region Non-public members

        private readonly ushort _listenerPort;

        private readonly ushort _maxWorkerThreadCount;

        private Socket _listenerSocket;

        private readonly Queue _requestQueue = new Queue();

        private ArrayList _workerThreadPool = new ArrayList();

        private Thread _listenerThread;

        private void WorkerThreadMethod()
        {
            int retryCounter = 0;
            while (_requestQueue.Count > 0 || retryCounter < 10)
            {
                Socket requestSocket = null;
                lock (_requestQueue)
                {
                    if (_requestQueue.Count > 0)
                    {
                        requestSocket = (Socket)_requestQueue.Dequeue();
                        retryCounter = 0;
                    }
                    else
                    {
                        retryCounter++;
                    }
                }

                if (requestSocket != null)
                {
                    if (requestSocket.Available > 0)
                    {
                        ProcessRequest(requestSocket);
                    }
                    requestSocket.Close();
                }
                Thread.Sleep(200);
            }

            lock (_workerThreadPool)
            {
                _workerThreadPool.Remove(Thread.CurrentThread);
            }
        }

        private void ListenerThreadMethod()
        {
            while (true)
            {
                Socket requestSocket;
                try
                {
                    requestSocket = _listenerSocket.Accept();
                    DebugLogger.GetLogger().Blink();

                    if (requestSocket == null)
                    {
                        continue;
                    }
                }
                catch (Exception)
                {
                    continue;
                }
                lock (_requestQueue)
                {
                    _requestQueue.Enqueue(requestSocket);
                }
                if (_workerThreadPool.Count < _maxWorkerThreadCount)
                {
                    lock (_workerThreadPool)
                    {
                        if (_workerThreadPool.Count < _maxWorkerThreadCount)
                        {
                            var workerThread = new Thread(WorkerThreadMethod);
                            _workerThreadPool.Add(workerThread);
                            workerThread.Start();
                        }
                    }
                }
            }
        }

        protected int WorkerThreadCount { get { return _workerThreadPool.Count; } }

        protected abstract void ProcessRequest(Socket pRequestSocket);

        #endregion

        #region Public members

        public bool IsAlive
        {
            get 
            { 
                var rVal = false;
                if (_listenerThread != null)
                {
                    rVal = _listenerThread.IsAlive;
                    if (rVal)
                    {
                        lock (_workerThreadPool)
                        {
                            foreach (var entry in _workerThreadPool)
                            {
                                var wt = (Thread)entry;
                                if (wt.IsAlive == false)
                                {
                                    rVal = false;
                                    break;
                                }
                            } 
                        }                        
                    }
                }
                return rVal;
            }
        }

        #region IDisposable members

        protected virtual void DisposeWorker() { }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool pDisposing)
        {
            if (pDisposing)
            {
                StopListening();
                DisposeWorker();
            }
        }

        public bool IsListening { get; private set; }

        #endregion

        public void StartListening()
        {
            //DebugHelper.Print("Socket server starting");
            _listenerSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            _listenerSocket.Bind(new IPEndPoint(IPAddress.Any, _listenerPort));
            //DebugHelper.Print("Listener socket bound to port: " + _listenerPort);
            _listenerSocket.Listen(30);
            //DebugHelper.Print("Socket listening");

            _listenerThread = new Thread(ListenerThreadMethod);
            //_startTime = DateTime.Now;
            _listenerThread.Start();
            //DebugHelper.Print("Listener thread started");
            IsListening = true;
        }

        public void StopListening()
        {
            //DebugHelper.Print("Server stopping");
            if (_listenerThread != null)
            {
                try
                {
                    _listenerThread.Abort();
                }
                catch (ThreadAbortException) { }
                _listenerThread = null;
            }

            IsListening = false;

            if (_workerThreadPool != null)
            {
                lock (_workerThreadPool)
                {
                    for (int i = 0; i < _workerThreadPool.Count; i++)
                    {
                        try
                        {
                            ((Thread)(_workerThreadPool[i])).Abort();
                        }
                        catch (ThreadAbortException) { }
                        _workerThreadPool[i] = null;
                    }
                }

                _workerThreadPool = null;
            }

            if (_listenerSocket != null)
            {
                _listenerSocket.Close();
                _listenerSocket = null;
            }

            if (_requestQueue != null && _requestQueue.Count > 0)
            {
                while (_requestQueue.Count > 0)
                {
                    var requestSocket = (Socket)_requestQueue.Dequeue();
                    requestSocket.Close();
                }
            }
        }

        #endregion

        #region Constuctors

        protected SocketServerBase(ushort pListenerPort, ushort pMaxWorkerThreadCount)
        {
            IsListening = false;
            _listenerPort = pListenerPort;
            _maxWorkerThreadCount = pMaxWorkerThreadCount;
        }

        ~SocketServerBase()
        {
            Dispose(false);
        }

        #endregion
    }
}
