﻿using System;
using System.Net;
using System.Net.Sockets;
using Core.Extensions;

namespace EmbeddedWebserver.Core.Internal
{
    public static class HttpRequestParser
    {
        #region Non-public members

        private const int MaxBufferedRequestLength = 1024;

        private static readonly char[] QueryStringTokenSeparator = new[] { '=' };

        private static readonly char[] RequestTokenSeparator = new[] { ' ' };

        private static readonly char[] HeaderTokenSeparator = new[] { ':' };

        private static HttpMethods ParseMethod(string pMethodString)
        {
            switch (pMethodString)
            {
                case "OPTIONS":
                    return HttpMethods.OPTIONS;
                case "GET":
                    return HttpMethods.GET;
                case "HEAD":
                    return HttpMethods.HEAD;
                case "POST":
                    return HttpMethods.POST;
                case "PUT":
                    return HttpMethods.PUT;
                case "DELETE":
                    return HttpMethods.DELETE;
                case "TRACE":
                    return HttpMethods.TRACE;
                case "CONNECT":
                    return HttpMethods.CONNECT;
                default:
                    throw new HttpException(HttpErrorCodes.HttpRequestParseError);
            }
        }

        private static void ParseRequestParameters(string pParameterString, StringDictionary pTargetCollection, IHttpServerUtility pServerUtility, bool pHttpDecode)
        {
            if (pTargetCollection == null)
            {
                throw new ArgumentNullException("pTargetCollection");
            }
            if (pServerUtility == null)
            {
                throw new ArgumentNullException("pServerUtility");
            }
            if (pParameterString.IsNullOrEmpty())
            {
                return;
            }
            string[] tokens = pParameterString.Split('&');
            foreach (string token in tokens)
            {
                string[] tokenParts = token.Split(QueryStringTokenSeparator, 2);
                string value = tokenParts.Length == 2 ? tokenParts[1] : null;
                if (value != null && pHttpDecode)
                {
                    value = pServerUtility.HtmlDecode(value);
                }
                pTargetCollection.Add(tokenParts[0], value);
            }
        }

        public static StringDictionary ParseParameters(string pParameterString, IHttpServerUtility pServerUtility)
        {
            var retval = new StringDictionary();
            ParseRequestParameters(pParameterString, retval, pServerUtility, true);
            return retval;
        }

        #endregion

        #region Public members

        public const string HeaderContentType = "Content-Type";

        public static HttpRequest ParseRequest(Socket pRequestSocket, IHttpServerUtility pServerUtility)
        {
            if (pServerUtility == null)
            {
                throw new ArgumentNullException("pServerUtility");
            }

            // Get request
            var bytesReceived = pRequestSocket.Available;
            var doBuffered = (bytesReceived <= MaxBufferedRequestLength);
            var retval = new HttpRequest(pRequestSocket, doBuffered);

            // User host
            var clientIP = pRequestSocket.RemoteEndPoint as IPEndPoint;
            if (clientIP != null) retval.UserHostAddress = clientIP.Address.ToString();

            // Get read buffer
            var readBuffer = new byte[MaxBufferedRequestLength];
            int readBufferLength = pRequestSocket.Receive(readBuffer, readBuffer.Length, SocketFlags.None);

            using (var requestStreamReader = new StringReader(readBuffer, readBufferLength))
            {
                // Read method + url + query string
                string currentLine = requestStreamReader.ReadLine();

                string[] splitLine = currentLine.Split(RequestTokenSeparator, 3);
                retval.Method = ParseMethod(splitLine[0]);

                string fullUrl = splitLine[1].Trim();
                string trimmedUrl = fullUrl;
                int queryStartIndex = trimmedUrl.IndexOf('?');
                if (queryStartIndex > 0)
                {
                    trimmedUrl = trimmedUrl.Substring(0, queryStartIndex);
                    ParseRequestParameters(fullUrl.Substring(queryStartIndex + 1), retval.QueryString, pServerUtility, false);
                }
                retval.Url = pServerUtility.UrlDecode(trimmedUrl.Trim('/').ToLower());

                // Read headers
                while (! (currentLine = requestStreamReader.ReadLine()).IsNullOrEmpty())
                {
                    splitLine = currentLine.Split(HeaderTokenSeparator, 2);
                    string headerName = splitLine[0].Trim();
                    string headerValue = splitLine[1].Trim();
                    retval.RequestHeaders.Add(headerName, headerValue);
                }

                if (doBuffered)
                {
                    // Read content
                    retval.RequestBody = requestStreamReader.ReadToEnd();
                }
                else
                {
                    retval.BytesPrefix = requestStreamReader.ReadRemainingBytes();
                }
            }

            return retval;
        }

        #endregion
    }
}
