﻿using System;
using System.Collections;
using System.IO;
using System.Net.Sockets;
using Core.Extensions;
using EmbeddedWebserver.Core.Configuration;
using EmbeddedWebserver.Core.Handlers;
using EmbeddedWebserver.Core.Handlers.Interfaces;
using EmbeddedWebserver.Core.Internal;
using EmbeddedWebserver.Core.Internal.Abstract;

namespace EmbeddedWebserver.Core
{
    public sealed class Webserver : SocketServerBase, IHttpServerUtility
    {
        #region Non-public members

        private const string PoweredByHeader = "Netduino";

        private static string MapStatusCodeToReason(HttpStatusCodes pStatusCode)
        {
            string retval = "";
            switch (pStatusCode)
            {
                case HttpStatusCodes.OK:
                    retval = "OK";
                    break;
                case HttpStatusCodes.Redirect:
                    retval = "Found";
                    break;
                case HttpStatusCodes.BadRequest:
                    retval = "Bad request";
                    break;
                case HttpStatusCodes.UnAuthorized:
                    retval = "Unauthorized";
                    break;
                case HttpStatusCodes.Forbidden:
                    retval = "Forbidden";
                    break;
                case HttpStatusCodes.NotFound:
                    retval = "Not found";
                    break;
                case HttpStatusCodes.MethodNotAllowed:
                    retval = "Method not allowed";
                    break;
                case HttpStatusCodes.InternalServerError:
                    retval = "Internal server error";
                    break;
                case HttpStatusCodes.ServiceUnavailable:
                    retval = "Service unavailable";
                    break;
            }
            return retval;
        }

        private static readonly FileSystemHandler DefaultHandler = new FileSystemHandler();

        private readonly Hashtable _handlers = new Hashtable();

        private IHandler ResolveHandler(HttpContext pContext)
        {
            if (FileSystemHandler.CanHandleRequest(pContext))
            {
                return DefaultHandler;
            }
            string targetUri = pContext.Request.Url;
            if (_handlers.Contains(targetUri))
            {
                object handlerItem = _handlers[targetUri];
                if (handlerItem is Type)
                {
                    return (IHandler)Activator.CreateInstance((Type)_handlers[targetUri]);
                }
                return (IHandler)handlerItem;
            }
            return null;
        }

        private void RegisterHandler(string pAccessUri, object pHandlerItem)
        {
            if (pAccessUri.IsNullOrEmpty())
            {
                throw new ArgumentNullException("pAccessUri");
            }
            if (IsListening)
            {
                throw new HttpException(HttpErrorCodes.ChangesNotAllowedWhileListening);
            }

            _handlers.Add(pAccessUri.Trim('/').ToLower(), pHandlerItem);
        }

        protected override void DisposeWorker()
        {

        }

        protected override void ProcessRequest(Socket pRequestSocket)
        {
            
            HttpRequest request = null;
            bool badRequest = false;
            try
            {
                request = HttpRequestParser.ParseRequest(pRequestSocket, this);
                if (request.Url.Equals(""))
                {
                    request.Url = "Index.htm";
                }
            }
            catch (Exception)
            {
                badRequest = true;
            }
            finally
            {
                if (badRequest)
                {
                    request = new HttpRequest(pRequestSocket);
                }
            }
            var context = new HttpContext(request, this);
            context.Request.Context = context;
            context.Response.Context = context;

            if (badRequest)
            {
                context.Response.StatusCode = HttpStatusCodes.BadRequest;
                context.Response.ResponseBody = null;
            }
            else
            {
                try
                {
                    IHandler handler = ResolveHandler(context);
                    if (handler == null)
                    {
                        context.Response.StatusCode = HttpStatusCodes.NotFound;
                    }
                    else
                    {
                        handler.ProcessRequest(context);
                    }
                }
                catch (Exception ex)
                {
                    context.Response.StatusCode = HttpStatusCodes.InternalServerError;
                    context.Response.ResponseBody = null;
                }
            }

            // Response header
            var headerBuilder = new StringBuilder("HTTP/1.1 ", 200);
            headerBuilder.Append(((int)context.Response.StatusCode).ToString());
            headerBuilder.Append(" ");
            headerBuilder.AppendLine(MapStatusCodeToReason(context.Response.StatusCode));
            context.Response.ResponseHeaders[HttpRequestParser.HeaderContentType] = context.Response.ContentType;
            bool hasResponse = context.Response.HasResponse;
            long responseLength = 0;
            byte[] sendBuffer = null;
            if (hasResponse)
            {
                if (context.Response.BufferedResponse)
                {
                    sendBuffer = System.Text.Encoding.UTF8.GetBytes(context.Response.ResponseBody);
                }
                var length = 0;
                if (sendBuffer != null)
                    length = sendBuffer.Length;
                responseLength = context.Response.BufferedResponse ? length : context.Response.ResponseStream.Length;
            }
            context.Response.ResponseHeaders["Content-Length"] = responseLength.ToString();
            context.Response.ResponseHeaders["Connection"] = "close";
            context.Response.ResponseHeaders["X-Powered-By"] = PoweredByHeader;

            if (context.Response.ResponseHeaders.Count > 0)
            {
                foreach (var responseHeaderKey in context.Response.ResponseHeaders.Keys)
                {
                    headerBuilder.Append((string)responseHeaderKey);
                    headerBuilder.Append(": ");
                    headerBuilder.AppendLine(context.Response.ResponseHeaders[(string)responseHeaderKey]);
                }
            }
            headerBuilder.AppendLine();
            pRequestSocket.Send(System.Text.Encoding.UTF8.GetBytes(headerBuilder.ToString()), headerBuilder.Length, SocketFlags.None);

            // Response content
            if (hasResponse)
            {
                // Buffered response
                if (context.Response.BufferedResponse)
                {
                    if (sendBuffer != null) pRequestSocket.Send(sendBuffer, sendBuffer.Length, SocketFlags.None);
                }
                // Streamed response                
                else
                {
                    sendBuffer = new byte[512];
                    int sendBytes;
                    while ((sendBytes = context.Response.ResponseStream.Read(sendBuffer, 0, sendBuffer.Length)) > 0)
                    {
                        pRequestSocket.Send(sendBuffer, sendBytes, SocketFlags.None);
                    }
                    context.Response.DisposeResponseStream();
                }
            }

            //Microsoft.SPOT.Debug.GC(true);
        }

        #endregion

        #region Public members

        #region IHttpServerUtility members

        public IEmbeddedWebapplicationConfiguration ApplicationConfiguration
        {
            get;
            private set;
        }

        public string MapPath(string pRelativePath)
        {
            return Path.Combine(ApplicationConfiguration.Path, pRelativePath.Replace("/", "\\"));
        }

        public string MapContentType(string pExtension)
        {
            String retval = "application/octet-stream";
            if (!pExtension.IsNullOrEmpty())
            {
                switch (pExtension.ToLower())
                {
                    case ".htm":
                    case ".html":
                        retval = "text/html";
                        break;
                    case ".js":
                        retval = "text/javascript";
                        break;
                    case ".css":
                        retval = "text/css";
                        break;
                    case ".gif":
                        retval = "image/gif";
                        break;
                    case ".jpg":
                        retval = "image/jpeg";
                        break;
                    case ".png":
                        retval = "image/png";
                        break;
                    case ".bmp":
                        retval = "image/bmp";
                        break;
                }
            }
            return retval;
        }

        public string UrlEncode(string pSourceString)
        {
            return pSourceString.Replace(" ", "%20");
        }

        public string UrlDecode(string pSourceString)
        {
            return pSourceString.Replace("%20", " ");
        }

        public string HtmlEncode(string pSourceString)
        {
            string retval = pSourceString.Replace("/", "%2F");
            retval = retval.Replace(" ", "+");
            return retval;
        }

        public string HtmlDecode(string pSourceString)
        {
            string retval = pSourceString.Replace("%2F", "/");
            retval = retval.Replace("+", " ");
            return retval;
        }

        #endregion

        public void RegisterHandler(string pAccessUri, Type pHandlerType)
        {
            if (pHandlerType == null)
            {
                throw new ArgumentNullException("pHandler");
            }

            bool validHandlerType = false;
            Type parentType = pHandlerType;
            while (parentType != null)
            {
                foreach (Type interfaceType in parentType.GetInterfaces())
                {
                    if (interfaceType == typeof(IHandler))
                    {
                        validHandlerType = true;
                        break;
                    }
                }
                if (validHandlerType)
                {
                    break;
                }
                parentType = parentType.BaseType;
            }

            if (!validHandlerType)
            {
                throw new ArgumentException("Invalid handler", "pHandlerType");
            }
            RegisterHandler(pAccessUri, (object) pHandlerType);
        }

        public void RegisterThreadSafeHandler(string pAccessUri, IHandler pHandlerInstance)
        {
            if (pHandlerInstance == null)
            {
                throw new ArgumentNullException("pHandlerInstance");
            }
            RegisterHandler(pAccessUri, pHandlerInstance);
        }

        #endregion

        #region Constructors

        public Webserver(IEmbeddedWebapplicationConfiguration pConfiguration)
            : base(pConfiguration.Port, pConfiguration.MaxWorkerThreadCount)
        {
            ApplicationConfiguration = pConfiguration;
        }

        #endregion
    }
}
