using System;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using ModBus.IO;
using ModBus.IO.Protocols;
using SolarDevices.Configuration;

namespace SolarDevices
{
    public class ModbusDevice : IDisposable
    {
        private const int POLL_PERIOD = 6000;
        private const int CountDownToDisconnect = 3;
        private int _lastUpdate = CountDownToDisconnect;

        private Timer _timer;
        private static ModbusDevice _singleton;
        private ushort[] _data;
        private Socket _socket;
        private ModbusClient _driver;
        private ModbusCommand _command;
        private ICommClient _portClient;
        private readonly IModbusConfiguration _configuration;
        private bool _connected;
        private bool _timeSet;
        private readonly Watchdog _watchdog;

        public ModbusDevice()
        {
            _configuration = new ModbusConfiguration();
            _watchdog = Watchdog.GetSingleton();
            _watchdog.RegisterTarget(Device.Modbus);
        }

        public void Dispose()
        {
            StopPolling();
            _singleton = null;
        }

        public static ModbusDevice GetSingleton()
        {
            return _singleton ?? (_singleton = new ModbusDevice());
        }

        public bool SetClassicDateTime()
        {
            var rVal = false;
            if (_driver != null)
            {
                var dt = DateTime.Now;
                if (dt.Year > 2012)
                {
                    if (UnlockClassic())
                    {
                        var dayOfWeek = (ushort)(dt.DayOfWeek);
                        var w1 = (ushort)(dt.Second & 0x003F);
                        w1 |= (ushort)((dt.Minute & 0x003F) << 8);

                        var w2 = (ushort)((dt.Hour & 0x001F));
                        w2 |= (ushort)(dayOfWeek << 8);

                        var w3 = (ushort)(dt.Day & 0x001F);
                        w3 |= (ushort)((dt.Month & 0x000F) << 8);

                        var w4 = (ushort)((dt.Year & 0x0FFF));

                        var w5 = (ushort)(dt.DayOfYear & 0x0FFF);

                        const int dataLength = 6;
                        var command = new ModbusCommand(ModbusCommand.FuncWriteMultipleRegisters) { Offset = 4213, Count = dataLength, Data = new ushort[dataLength] };
                        command.Data[0] = w1;
                        command.Data[1] = w2;
                        command.Data[2] = w3;
                        command.Data[3] = w4;
                        command.Data[4] = w5;
                        command.Data[5] = 0;
                        var result = _driver.ExecuteGeneric(_portClient, command);
                        rVal = result.Status == CommResponse.Ack;
                        //if (rVal)
                        //{
                        //    var readCommand = new ModbusCommand(ModbusCommand.FuncReadMultipleRegisters) { Offset = 4213, Count = dataLength };
                        //    _driver.ExecuteGeneric(_portClient, readCommand);
                        //    var result2 = _driver.ExecuteGeneric(_portClient, _command);
                        //    if (result2.Status == CommResponse.Ack)
                        //    {
                        //        for (int i = 0; i < dataLength; i++)
                        //        {
                        //            SolarSystem.TryLog("Classic time [" + i + "] wrote: " + command.Data[i] + " read: " + readCommand.Data[i]);
                        //        }
                        //    }
                        //}
                    }
                }
            }
            return rVal;
        }

        private bool UnlockClassic()
        {
            if (_configuration.ModbusPassword == null || _configuration.ModbusPassword == string.Empty )
            {
                return true; // assume password jumper was pulled
            }
            const int dataLength = 2;
            const int offset = 20491; // or possibly 28672
            var writeCommand = new ModbusCommand(ModbusCommand.FuncWriteMultipleRegisters) { Offset = offset, Count = dataLength, Data = new ushort[dataLength] };
            var password = UInt32.Parse(_configuration.ModbusPassword);
            writeCommand.Data[0] = (ushort)((password >> 16) & 0xffff);
            writeCommand.Data[1] = (ushort)(password & 0xffff); ;
            var writeResult = _driver.ExecuteGeneric(_portClient, writeCommand);
            var rVal = writeResult.Status == CommResponse.Ack;
            return rVal;
        }

        public IModbusConfiguration Configuration
        {
            get { return _configuration; }
        }

        public void StartPolling()
        {
            if (_timer == null)
                SetupTimer();
        }

        private void SetupTimer()
        {
            _timer = new Timer(PollDevice, null, 1000, POLL_PERIOD);
        }

        public void StopPolling()
        {
            try
            {
                if (_timer != null)
                {
                    _timer.Dispose();
                }
                if (_socket != null)
                {
                    _socket.Close();
                }
            }
            catch (Exception)
            {

            }
            _timer = null;
            _socket = null;
            _portClient = null;
            _driver = null;
            _command = null;
        }

        private void PollDevice(object state)
        {
            try
            {

                if (_socket == null)
                {
                    _socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                    _socket.SetSocketOption(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, true);
                    _socket.SendTimeout = 2000;
                    _socket.ReceiveTimeout = 2000;
                    _lastUpdate = CountDownToDisconnect;
                }
                if (_portClient == null)
                {
                    try
                    {
                        var ipaddr = IPAddress.Parse(Configuration.ModbusSlaveIp);
                        var ept = new IPEndPoint(ipaddr, Configuration.ModbusSlavePort);
                        _socket.Connect(ept);
                        _portClient = _socket.GetClient();
                        _driver = new ModbusClient(new ModbusTcpCodec()) { Address = 1 };
                        _command = new ModbusCommand(ModbusCommand.FuncReadMultipleRegisters) { Offset = Address, Count = Count };
                        SolarSystem.TryLog("Connected to modbus device");
                        _watchdog.CallIn(Device.Modbus);
                    }
                    catch (Exception ex)
                    {
                        lock (this)
                        {
                            _connected = false;
                        }
                        _socket = null;
                        _portClient = null;
                        SolarSystem.TryLog(ex);
                        return;
                    }
                }
                if (_driver != null)
                {
                    try
                    {
                        if (_timeSet == false && Configuration.SetClassicTime)
                        {
                            _watchdog.CallIn(Device.Modbus);
                            if (SetClassicDateTime())
                            {
                                _timeSet = true;
                                SolarSystem.TryLog("Updated classic date & time");
                            }
                        }
                        //execute the command synchronously
                        var result = _driver.ExecuteGeneric(_portClient, _command);
                        if (result.Status == CommResponse.Ack)
                        {
                            lock (this)
                            {
                                _connected = true;
                                _data = _command.Data;
                                _lastUpdate = CountDownToDisconnect;
                            }
                        }
                        else
                        {
                            lock (this)
                            {
                                _connected = false;
                            }
                            _socket = null;
                            _portClient = null;
                            _driver = null;
                            _command = null;
                            SolarSystem.TryLog("Disconnected from modbus device");
                        }
                        if (--_lastUpdate != 0)
                            _watchdog.CallIn(Device.Modbus);
                    }
                    catch (Exception ex)
                    {
                        lock (this)
                        {
                            _connected = false;
                        }
                        _socket = null;
                        _portClient = null;
                        _driver = null;
                        _command = null;
                        SolarSystem.TryLog(ex);
                    }
                }

            }
            catch (Exception ex) // catch all
            {
                SolarSystem.TryLog(ex);
            }
        }

        public int Address { get; set; }
        public int Count { get; set; }

        public ushort[] GetData()
        {
            ushort[] rval;
            lock (this)
            {
                if (_data != null)
                {
                    rval = new ushort[_data.Length];
                    _data.CopyTo(rval, 0);
                }
                else
                {
                    rval = null;
                }
            }
            return rval;
        }

        public bool Connected
        {
            get
            {
                bool rval;
                lock (this)
                {
                    rval = _connected;
                }
                return rval;
            }
        }

    }
}